from typing import Union, List
import numpy as np
import casadi as ca
import forwardkinematics.urdfFks.casadiConversion.urdfparser as u2c

from forwardkinematics.fksCommon.fk import ForwardKinematics


class LinkNotInURDFError(Exception):
    pass


class URDFForwardKinematics(ForwardKinematics):
    def __init__(self, urdf: str, root_link: str, end_links: List[str], base_type: str = 'holonomic'):
        super().__init__()
        self._urdf = urdf
        self._root_link = root_link
        self._end_links = end_links
        self._base_type = base_type
        self.read_urdf()
        self._n = self.robot.degrees_of_freedom()
        self._q_ca = ca.SX.sym("q", self._n)
        self._mount_transformation = np.identity(4)
        if base_type in ['diffdrive']:
            self._q_base = ca.SX.sym("q_base", 3)
        self.generate_functions()

    def n(self) -> int:
        if self._base_type == 'diffdrive':
            return self._n + 3
        return self._n

    def read_urdf(self):
        self.robot = u2c.URDFparser(root_link=self._root_link, end_links=self._end_links)
        self.robot.from_string(self._urdf)
        self.robot.detect_link_names()
        self.robot.set_joint_variable_map()

    def generate_functions(self):
        self._fks = {}
        for link in self.robot.link_names():
            if self._base_type in ['diffdrive']:
                q = ca.vcat([self._q_base, self._q_ca])
            else:
                q = self._q_ca
            ca_fun = ca.Function(
                "fk" + link, [q], [self.casadi(q, link)]
            )
            self._fks[link] = ca_fun

    def casadi(self, q: ca.SX, child_link: str, parent_link: Union[str, None] = None, link_transformation=np.eye(4), position_only=False):
        if parent_link is None:
            parent_link = self._root_link
        if child_link not in self.robot.link_names():
            raise LinkNotInURDFError(
                f"""The link you have requested, {child_link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if self._base_type in ['diffdrive']:
            fk = self.robot.get_forward_kinematics(parent_link, child_link, q[2:], link_transformation)["T_fk"]
            c = ca.cos(q[2])
            s = ca.sin(q[2])
            T_base = ca.vcat([
                ca.hcat([c, -s, 0, q[0]]),
                ca.hcat([s, c, 0, q[1]]),
                ca.hcat([0, 0, 1, 0]),
                ca.hcat([0, 0, 0, 1]),
            ])
            fk = ca.mtimes(T_base, fk)
        else:
            fk = self.robot.get_forward_kinematics(parent_link, child_link, q, link_transformation)["T_fk"]
            fk = ca.mtimes(self._mount_transformation, fk)

        if position_only:
            fk = fk[0:3, 3]
        return fk

    def numpy(self, q: ca.SX, child_link: str, parent_link: Union[str, None] = None, link_transformation=np.eye(4), position_only=False):
        if child_link not in self._fks and child_link != self._root_link:
            raise LinkNotInURDFError(
                f"""The link you have requested, {child_link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if parent_link is None:
            parent_link = self._root_link
        if parent_link not in self._fks and parent_link != self._root_link:
            raise LinkNotInURDFError(
                f"""The link you have requested, {parent_link}, is not in the urdf.
                    Possible links are  {self.robot.link_names()}"""
            )
        if parent_link == self._root_link:
            fk_parent = np.identity(4)
        else:
            fk_parent = self._fks[parent_link](q)
        if child_link == self._root_link:
            fk_child = np.identity(4)
        else:
            fk_child = self._fks[child_link](q)
        tf_parent_child = np.dot(np.linalg.inv(fk_parent), fk_child)
        tf_parent_child = np.dot(link_transformation, tf_parent_child) #ToDo check if correct
        if position_only:
            return tf_parent_child[0:3, 3]
        else:
            return tf_parent_child




from typing import Union, List, Optional
import logging
import xml.etree.ElementTree as ET
import numpy as np
import casadi as ca

from forwardkinematics.fksCommon.fk import ForwardKinematics

casadiOrNumpy = Union[ca.SX, np.ndarray]

def matrix_multiply(A: casadiOrNumpy, B: casadiOrNumpy) -> casadiOrNumpy:
    if isinstance(A, np.ndarray) and isinstance(B, np.ndarray):
        return np.dot(A, B)
    elif isinstance(A, ca.SX) and isinstance(B, ca.SX):
        return ca.mtimes(A, B)
    elif isinstance(A, ca.SX) and isinstance(B, np.ndarray):
        return ca.mtimes(A, B)
    elif isinstance(A, np.ndarray) and isinstance(B, ca.SX):
        return ca.mtimes(A, B)
    else:
        raise ValueError(f"A({type(A)}) and B({type(B)}) must be either casadi or numpy arrays.")

def get_T_element(
        element: ET.Element,
    ) -> np.ndarray:
    """
    Get the transformation matrix of an element.
    """
    pos = element.get('pos') if 'pos' in element.attrib else '0 0 0'
    offset = np.array([float(i) for i in pos.split()])
    if 'quat' in element.attrib:
        quat = [float(i) for i in element.attrib['quat'].split()]
        rotation_matrix = quat2rotMat(quat)
    else:
        rotation_matrix = np.eye(3)
    return homogeneous_transformation(
        rotation=rotation_matrix,
        translation=offset
    )

def quat2rotMat(quat: List[float]) -> np.ndarray:
    """
    Convert a quaternion to a rotation matrix.
    """
    # Normalize the quaternion
    quat_np = np.array(quat)
    q = quat_np / np.linalg.norm(quat_np)
    q0, q1, q2, q3 = q[0], q[1], q[2], q[3]
    R = np.array([
        [1 - 2 * q2 * q2 - 2 * q3 * q3, 2 * q1 * q2 - 2 * q0 * q3, 2 * q1 * q3 + 2 * q0 * q2],
        [2 * q1 * q2 + 2 * q0 * q3, 1 - 2 * q1 * q1 - 2 * q3 * q3, 2 * q2 * q3 - 2 * q0 * q1],
        [2 * q1 * q3 - 2 * q0 * q2, 2 * q2 * q3 + 2 * q0 * q1, 1 - 2 * q1 * q1 - 2 * q2 * q2]
    ])
    return R

def axis2rotMat(axis: List[int], theta: casadiOrNumpy) -> casadiOrNumpy:
    """
    Return the rotation matrix associated with counterclockwise rotation about
    the given axis by theta radians.
    """
    if isinstance(theta, np.float64) or isinstance(theta, float):
        axis = np.array(axis)
        axis = axis / np.linalg.norm(axis)
        a = np.cos(theta / 2.0)
        b, c, d = -axis * np.sin(theta / 2.0)
        aa, bb, cc, dd = a * a, b * b, c * c, d * d
        bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
        R = np.array([
            [aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)],
            [2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)],
            [2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc]
        ])
        return R
    elif isinstance(theta, ca.SX):
        axis_ca = ca.SX(axis)
        axis_ca = axis_ca / ca.sqrt(ca.dot(axis_ca, axis_ca))
        a = ca.cos(theta / 2.0)
        b, c, d = ca.vertsplit(-axis_ca * ca.sin(theta / 2.0))
        aa, bb, cc, dd = a * a, b * b, c * c, d * d
        bc, ad, ac, ab, bd, cd = b * c, a * d, a * c, a * b, b * d, c * d
        R = ca.vertcat(ca.horzcat(aa + bb - cc - dd, 2 * (bc + ad), 2 * (bd - ac)),
                       ca.horzcat(2 * (bc - ad), aa + cc - bb - dd, 2 * (cd + ab)),
                       ca.horzcat(2 * (bd + ac), 2 * (cd - ab), aa + dd - bb - cc))
        return ca.SX(R)
    else:
        raise ValueError(f"theta {type(theta)} must be either casadi or numpy array.")

def homogeneous_transformation(
        rotation: Optional[casadiOrNumpy] = None,
        translation: Optional[casadiOrNumpy] = None,
    ) -> casadiOrNumpy:
    """
    Construct a homogeneous transformation matrix from rotation matrix and translation vector.
    """
    if rotation is not None and isinstance(rotation, ca.SX) or translation is not None and isinstance(translation, ca.SX):
        T = ca.SX.eye(4)
    else:
        T = np.eye(4)
    if rotation is not None:
        T[:3, :3] = rotation
    if translation is not None:
        T[:3, 3] = translation
    return T


class GenericXMLFk(ForwardKinematics):
    """
    Forward kinematics class for Mujoco XML files.
    """
    def __init__(self, xml: str, root_link: str, end_links: List[str], base_type: str = 'holonomic'):
        """
        Initialize the forward kinematics class.

        Parameters:
            xml (str): the XML string.
        """
        super().__init__()
        self._xml = xml
        self.tree = ET.ElementTree(ET.fromstring(xml))
        self.root = self.tree.getroot()


    def casadi(
        self, q: ca.SX,
        child_link: str,
        parent_link: Union[str, None] = None,
        link_transformation: np.ndarray = np.eye(4),
        position_only: bool = False
    ) -> ca.SX:
        return self.fk(
            q, child_link, parent_link, link_transformation, position_only
        )

    def numpy(
        self, q: np.ndarray,
        child_link: str,
        parent_link: Union[str, None] = None,
        link_transformation: np.ndarray = np.eye(4),
        position_only: bool = False
    ) -> ca.SX:
        return self.fk(
            q, child_link, parent_link, link_transformation, position_only
        )

    def fk(
        self, q: Union[ca.SX, np.ndarray],
        child_link: str,
        parent_link: Union[str, None] = None,
        link_transformation: np.ndarray = np.eye(4),
        position_only: bool = False
    ) -> casadiOrNumpy:
        """
        Compute the forward kinematics of the robot.
        
        Parameters:
            q (ca.SX): The joint configuration.
            child_link (str): The name of the child link.
            parent_link (str): The name of the parent link.
            link_transformation (np.ndarray): The transformation matrix of the parent link.
            position_only (bool): If True, only the position of the end-effector is returned.
        
        """
        root = self.root.find('.//worldbody')
        T = homogeneous_transformation()
        joint_counter = 0
        child_link_found = False
        for element in root.iter():
            if child_link_found:
                break
            T_element = np.eye(4)
            if element.tag == 'body':
                T_element = get_T_element(element)
                if element.get('name') == child_link:
                    child_link_found = True
            if element.tag == 'joint':
                joint_type = 'hinge' if 'type' not in element.attrib else element.attrib['type']
                if 'axis' not in element.attrib:
                    joint_axis = [0, 0, 1]
                    logging.warning(f"Joint axis not found in XML. Assuming default axis {joint_axis}.")
                else:
                    joint_axis = [int(x) for x in element.attrib['axis'].split()]
                if joint_type == 'hinge':
                    T_element = homogeneous_transformation(
                        rotation=axis2rotMat(
                            joint_axis,
                            q[joint_counter],
                        )
                    )
                elif joint_type == 'slide':
                    T_element = homogeneous_transformation(
                        translation=joint_axis * q[joint_counter]
                    )
                joint_counter += 1
            if element.tag == 'geom':
                if element.get('name') == child_link:
                    T_element = get_T_element(element)
                    child_link_found = True
                if element.get('mesh') == child_link:
                    T_element = get_T_element(element)
                    child_link_found = True
            T = matrix_multiply(T, T_element)
        if not child_link_found:
            raise ValueError(f"Child link {child_link} not found in XML.")
        else:
            if position_only:
                return T[0:3,3]
            else:
                return T






"""This module is in most parts copied from https://github.com/mahaarbo/urdf2casadi.

Changes are in get_forward_kinematics as it allows to pass the variable as an argument.
"""
import casadi as ca
import numpy as np
from urdf_parser_py.urdf import URDF
import forwardkinematics.urdfFks.casadiConversion.geometry.transformation_matrix as T


class URDFparser(object):
    """Class that turns a chain from URDF to casadi functions."""
    actuated_types = ["prismatic", "revolute", "continuous"]
    
    def __init__(self, root_link: str="base_link", end_links: list = None):
        self._root_link = root_link
        if isinstance(end_links, str):
            self._end_links = [end_links]
        else:
            self._end_links = end_links

    def extract_information(self):
        self._actuated_joints = []
        self._active_joints = set()
        self._degrees_of_freedom = 0
        self.detect_link_names()
        self._absolute_root_link = self.robot_desc.get_root()
        self.set_active_joints()
        self.set_actuated_joints()
        self.extract_degrees_of_freedom()

    def from_file(self, filename):
        """Uses an URDF file to get robot description."""
        self.robot_desc = URDF.from_xml_file(filename)
        self.extract_information()

    def from_server(self, key="robot_description"):
        """Uses a parameter server to get robot description."""
        self.robot_desc = URDF.from_parameter_server(key=key)
        self.extract_information()

    def from_string(self, urdfstring):
        """Uses a URDF string to get robot description."""
        self.robot_desc = URDF.from_xml_string(urdfstring)
        self.extract_information()

    def set_actuated_joints(self) -> None:
        for joint in self.robot_desc.joints:
            if joint.type in self.actuated_types:
                self._actuated_joints.append(joint.name)

    def degrees_of_freedom(self) -> int:
        return self._degrees_of_freedom

    def extract_degrees_of_freedom(self) -> None:
        self._degrees_of_freedom = 0
        for joint_name in self._active_joints:
            if joint_name in self._actuated_joints:
                self._degrees_of_freedom += 1

    def actuated_joints(self) -> list:
        return self._actuated_joints

    def set_joint_variable_map(self) -> None:
        self._joint_map = {}
        index = 0
        for joint_name in self._actuated_joints:
            if joint_name in self._active_joints:
                self._joint_map[joint_name] = index
                index += 1

    def is_active_joint(self, joint):
        parent_link = joint.parent
        while parent_link not in [self._root_link, self._absolute_root_link]:
            if parent_link in self._end_links:
                return False
            parent_joint, parent_link = self.robot_desc.parent_map[parent_link]
            if parent_joint in self._active_joints:
                return True

        if parent_link == self._root_link:
            return True
        return False

    def set_active_joints(self) -> None:
        for parent_link in self._end_links:
            while parent_link not in [self._root_link, self._absolute_root_link]:
                parent_joint, parent_link = self.robot_desc.parent_map[parent_link]
                self._active_joints.add(parent_joint)
                if parent_link == self._root_link:
                    break
        
    def active_joints(self) -> set:
        return self._active_joints
        

    def get_joint_info(self, root, tip) -> list:
        """Using an URDF to extract joint information, i.e list of
        joints, actuated names and upper and lower limits."""
        chain = self.robot_desc.get_chain(root, tip)
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')

        joint_list = []

        for item in chain:
            if item in self.robot_desc.joint_map:
                joint = self.robot_desc.joint_map[item]
                if joint.name in self._active_joints:
                    joint_list += [joint]

        return joint_list

    def link_names(self):
        return self._link_names

    def detect_link_names(self):
        self._link_names = []
        for link in self.robot_desc.links:
            if link.name in self.robot_desc.parent_map:
                self._link_names.append(link.name)
            else:
                print(f"Link with name {link.name} does not has a parent. Link name is skipped.")
        return self._link_names

    def get_forward_kinematics(self, root, tip, q, link_transformation=np.eye(4)):
        """Returns the forward kinematics as a casadi function."""
        if self.robot_desc is None:
            raise ValueError('Robot description not loaded from urdf')
        joint_list = self.get_joint_info(self._absolute_root_link, tip)
        T_fk = ca.SX.eye(4)
        for joint in joint_list:
            if joint.type == "fixed":
                xyz = joint.origin.xyz
                rpy = joint.origin.rpy
                joint_frame = T.numpy_rpy(xyz, *rpy)
                T_fk = ca.mtimes(T_fk, joint_frame)

            elif joint.type == "prismatic":
                if joint.axis is None:
                    axis = ca.np.array([1., 0., 0.])
                else:
                    axis = ca.np.array(joint.axis)
                joint_frame = T.prismatic(joint.origin.xyz,
                                          joint.origin.rpy,
                                          joint.axis, q[self._joint_map[joint.name]])
                T_fk = ca.mtimes(T_fk, joint_frame)

            elif joint.type in ["revolute", "continuous"]:
                if joint.axis is None:
                    axis = ca.np.array([1., 0., 0.])
                else:
                    axis = ca.np.array(joint.axis)
                axis = (1./ca.np.linalg.norm(axis))*axis
                joint_frame = T.revolute(
                    joint.origin.xyz,
                    joint.origin.rpy,
                    joint.axis, q[self._joint_map[joint.name]])
                T_fk = ca.mtimes(T_fk, joint_frame)

        T_fk = ca.mtimes(T_fk, link_transformation)

        return {
            "T_fk": T_fk
        }



from typing import Union
import numpy as np
import casadi as ca

class ForwardKinematics():

    """Abstract class for forward kinematics"""

    _n: int

    def __init__(self):
        pass

    def set_mount_transformation(self, mount_transformation: np.ndarray):
        self._mount_transformation = mount_transformation

    def n(self) -> int:
        return self._n

    def casadi(
            self, q: ca.SX,
            child_link: Union[int, str],
            parent_link: Union[int, str, None] = None,
            link_transformation=np.eye(4),
            position_only: bool = False
        ) -> ca.SX:
        raise NotImplementedError

    def numpy(
            self, q: np.ndarray,
            child_link: Union[int, str],
            parent_link: Union[int, str, None] = None,
            link_transformation=np.eye(4),
            position_only: bool = False
        ) -> np.ndarray:
        raise NotImplementedError





from typing import Union
import numpy as np
import casadi as ca
from forwardkinematics.planarFks.planar_fk import ForwardKinematicsPlanar


def get_rotation_matrix_casadi(angle: ca.SX) -> ca.SX:
    print(angle)
    return ca.SX(
        ca.vcat(
            [
                ca.hcat([ca.cos(angle), -ca.sin(angle), 0]),
                ca.hcat([ca.sin(angle), ca.cos(angle), 0]),
                ca.hcat([0, 0, 1]),
            ]
        )
    )


def get_rotation_matrix_numpy(angle: ca.SX) -> np.ndarray:
    return np.array(
        [
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1],
        ]
    )


class PlanarArmFk(ForwardKinematicsPlanar):
    def __init__(self, n):
        super().__init__()
        self._n = n

    def casadi(
        self,
        q: ca.SX,
        child_link: Union[int, str],
        parent_link: Union[int, str, None] = None,
        link_transformation=np.eye(3),
        position_only: bool = False,
    ):
        child_link = self.ensure_int_link(child_link)
        if not parent_link:
            parent_link = 0
        parent_link = self.ensure_int_link(parent_link)
        if not parent_link:
            parent_link = 0
        fk = get_rotation_matrix_casadi(q[parent_link])
        for i in range(parent_link + 1, child_link + 1):
            if i == self.n():
                fk_i = get_rotation_matrix_casadi(0)
            else:
                fk_i = get_rotation_matrix_casadi(q[i])
            fk_i[0, 2] = 1
            fk = ca.mtimes(fk, fk_i)
        fk = ca.mtimes(fk, link_transformation)
        if position_only:
            return fk[0:2, 2]
        else:
            return fk

    def numpy(
        self,
        q: np.ndarray,
        child_link: Union[int, str],
        parent_link: Union[int, str, None] = None,
        link_transformation=np.eye(3),
        position_only: bool = False,
    ):
        child_link = self.ensure_int_link(child_link)
        if not parent_link:
            parent_link = 0
        parent_link = self.ensure_int_link(parent_link)
        fk = get_rotation_matrix_numpy(q[parent_link])
        for i in range(parent_link + 1, child_link + 1):
            if i == self.n():
                fk_i = get_rotation_matrix_numpy(0)
            else:
                fk_i = get_rotation_matrix_numpy(q[i])
            fk_i[0, 2] = 1
            fk = np.dot(fk, fk_i)
        fk = np.dot(fk, link_transformation)
        if position_only:
            return fk[0:2, 2]
        else:
            return fk



from typing import Union

import numpy as np
import casadi as ca

from forwardkinematics.planarFks.planar_fk import ForwardKinematicsPlanar


class PointFk(ForwardKinematicsPlanar):
    def __init__(self):
        super().__init__()
        self._n = 2

    def casadi(
        self,
        q: ca.SX,
        child_link: Union[int, str],
        parent_link: Union[int, str, None] = None,
        link_transformation=np.eye(3),
        position_only: bool = False,
    ):
        fk = ca.SX.eye(3)
        child_link = self.ensure_int_link(child_link)
        if not parent_link:
            parent_link = 0
        parent_link = self.ensure_int_link(parent_link)
        if not parent_link:
            parent_link = 0
        if parent_link == 0 and child_link > 0:
            fk[0:2, 2] = q
        fk = ca.mtimes(fk, link_transformation)
        if position_only:
            return fk[0:2, 2]
        else:
            return fk

    def numpy(
        self,
        q: np.ndarray,
        child_link: Union[int, str],
        parent_link: Union[int, str, None] = None,
        link_transformation=np.eye(3),
        position_only: bool = False,
    ):
        fk = np.eye(3)
        child_link = self.ensure_int_link(child_link)
        if not parent_link:
            parent_link = 0
        parent_link = self.ensure_int_link(parent_link)
        if not parent_link:
            parent_link = 0
        if parent_link == 0 and child_link > 0:
            fk[0:2, 2] = q
        fk = np.dot(fk, link_transformation)
        if position_only:
            return fk[0:2, 2]
        else:
            return fk



from importlib.metadata import version
__version__ = version(__name__)

from forwardkinematics.urdfFks.generic_urdf_fk import GenericURDFFk
from forwardkinematics.xmlFks.generic_xml_fk import GenericXMLFk



from typing import Union
import re

import numpy as np
import casadi as ca

from forwardkinematics.fksCommon.fk import ForwardKinematics

class NoLinkIndexFoundInLinkNameError(Exception):
    pass

class ForwardKinematicsPlanar(ForwardKinematics):

    def __init__(self):
        super().__init__()
        self._mount_transformation = np.identity(3)

    def ensure_int_link(self, link: Union[str, int]) -> int:
        if isinstance(link, int):
            return link
        regex_match = re.match(r'\D*(\d*)', link)
        try:
            return int(regex_match.group(1))
        except Exception as _:
            raise NoLinkIndexFoundInLinkNameError(f"Link name {link} could not be resolved")

    def casadi(
            self, q: ca.SX,
            child_link: Union[int, str],
            parent_link: Union[int, str, None] = None,
            link_transformation=np.eye(3),
            position_only: bool = False
        ) -> ca.SX:
        raise NotImplementedError

    def numpy(
            self, q: np.ndarray,
            child_link: Union[int, str],
            parent_link: Union[int, str, None] = None,
            link_transformation=np.eye(3),
            position_only: bool = False
        ) -> np.ndarray:
        raise NotImplementedError
